/* Copyright (c) 2015-present, salesforce.com, inc. All rights reserved
  Licensed under BSD 3-Clause - see LICENSE.txt or git.io/sfdc-license */

import ReactiveElement from 'elix/src/core/ReactiveElement';
import { templateFrom } from 'elix/src/core/htmlLiterals';
import FocusVisibleMixin from 'elix/src/base/FocusVisibleMixin';
import { template, render, state, firstRender, setState, defaultState } from 'elix/src/base/internal';
import { standardBooleanAttributes } from 'elix/src/core/AttributeMarshallingMixin';
import { nanoid } from 'nanoid';

import common from '@salesforce-ux/sds-common/common.module.css';
import html from './input-checkbox.html';
import css from './input-checkbox.module.css';

/**
 * Extend Elix's standard boolean attributes definition
 */
Object.assign(standardBooleanAttributes, {
  required: true,
  indeterminate: true,
  autofocus: true,
});

export default class SdsInputCheckbox extends FocusVisibleMixin(ReactiveElement) {
  get [defaultState]() {
    return Object.assign(super[defaultState], {
      indicator: 'check',
      indeterminateIndicator: 'dash',
      id: `input-checkbox-${nanoid(10)}`,
    });
  }

  get checked() {
    return this[state].checked;
  }

  set checked(checked) {
    this[setState]({ checked });
  }

  get indeterminate() {
    return this[state].indeterminate;
  }

  set indeterminate(indeterminate) {
    this[setState]({ indeterminate });
  }

  get disabled() {
    return this[state].disabled;
  }

  set disabled(disabled) {
    this[setState]({ disabled });
  }

  get required() {
    return this[state].required;
  }

  set required(required) {
    this[setState]({ required });
  }

  get name() {
    return this[state].name;
  }

  set name(name) {
    this[setState]({ name });
  }

  get value() {
    return this[state].value;
  }

  set value(value) {
    this[setState]({ value });
  }

  get id() {
    return this[state].id;
  }

  set id(id) {
    this[setState]({ id });
  }

  get indicator() {
    return this[state].indicator;
  }

  set indicator(indicator) {
    this[setState]({ indicator });
  }

  get indeterminateIndicator() {
    return this[state].indeterminateIndicator;
  }

  set indeterminateIndicator(indeterminateIndicator) {
    this[setState]({ indeterminateIndicator });
  }

  get error() {
    return this[state].error;
  }

  set error(error) {
    this[setState]({ error });
  }

  /**
   * Handle internal 'change' event triggered by checkbox
   */
  _handleChange(checkbox) {
    /**
     * Sync our states with the IDL properties of the checkbox
     */
    this[setState]({ checked: checkbox.checked });

    /**
     * Indeterminate state can only be set through a content attribute or IDL
     * property. Clicking an indeterminate checkbox updates the checkbox's
     * checked IDL property to true. Since the checkbox is now checked,
     * interminate is removed and not active until set another element.
     */
    if (this[state].indeterminate) {
      this[setState]({ indeterminate: false });
    }

    dispatchCustomEvent(this, 'change');
  }

  /**
   * Handle internal 'focus' event triggered by checkbox
   */
  _handleFocus() {
    dispatchCustomEvent(this, 'focus');
  }

  /**
   * Handle internal 'blur' event triggered by checkbox
   */
  _handleBlur() {
    dispatchCustomEvent(this, 'blur');
  }

  /**
   * Check to make sure attributes are not being set incorrectly.
   */
  _validateAttributes() {
    if (this[state].checked && this[state].indeterminate) {
      throw Error(
        'The sds-input-checkbox can only be either checked or indeterminate; it cannot be both at the same time.',
      );
    }
  }

  [render](changed) {
    super[render](changed);

    const label = this.shadowRoot.querySelector('[part~="label"]');
    const checkbox = this.shadowRoot.querySelector('[part~="checkbox"');
    const icon = this.shadowRoot.querySelector('sds-icon');
    const { checked, indeterminate, required, disabled, name, value, id, indicator, indeterminateIndicator } =
      this[state];

    if (this[firstRender] && checkbox) {
      checkbox.addEventListener('change', (event) => {
        this._handleChange(checkbox, event);
      });

      checkbox.addEventListener('focus', (event) => {
        this._handleFocus(event);
      });

      checkbox.addEventListener('blur', (event) => {
        this._handleBlur(event);
      });

      /**
       * If present, set the default, one-time "checked" attribute
       */
      if (this.checked) {
        checkbox.setAttribute('checked', '');
      }
    }

    /**
     * Toggle the 'disabled' attribute on the host element and sync checkbox's 'disabled' content attribute to state
     */
    if (changed.disabled) {
      toggleBooleanAttribute(this, 'disabled', disabled);
      toggleBooleanAttribute(checkbox, 'disabled', disabled);
    }

    /**
     * Toggle the 'required' attribute on the host element and sync checkbox's 'required' content attribute to state
     */
    if (changed.required) {
      toggleBooleanAttribute(this, 'required', required);
      toggleBooleanAttribute(checkbox, 'required', required);
    }

    /**
     * Set the icon for the indicator
     */
    if (changed.indicator) {
      icon.setAttribute('symbol', indicator);
    }

    /**
     * Set the icon for the indeterminate indicator
     */
    if (changed.indeterminateIndicator) {
      indeterminate && icon.setAttribute('symbol', indeterminateIndicator);
    }

    /**
     * Toggle the 'checked' attribute on the host element and sync checkbox's 'checked' IDL property to state
     */
    if (changed.checked) {
      toggleBooleanAttribute(this, 'checked', checked);
      checkbox.checked = checked;

      checked && icon.setAttribute('symbol', indicator);
    }

    /**
     * Toggle the 'indeterminate' attribute on the host element and sync checkbox's 'indeterminate' IDL property to state
     */
    if (changed.indeterminate) {
      toggleBooleanAttribute(this, 'indeterminate', indeterminate);
      checkbox.indeterminate = indeterminate;

      indeterminate && icon.setAttribute('symbol', indeterminateIndicator);
      indeterminate && icon.setAttribute('prefetch', indicator);
    }

    /**
     * Delegate content 'name' attribute and value to the checkbox
     */
    if (changed.name) {
      if (name) {
        delegateAttribute(this, checkbox, 'name', name);
      }
    }

    /**
     * Delegate content 'value' attribute and value to the checkbox
     */
    if (changed.value) {
      if (value) {
        delegateAttribute(this, checkbox, 'value', value);
      }
    }

    /**
     * Delegate content 'value' attribute and value to the checkbox
     */
    if (changed.id) {
      if (id) {
        delegateAttribute(this, checkbox, 'id', id);
        label.setAttribute('for', id);
      }
    }

    if (changed.checked || changed.indeterminate) {
      this._validateAttributes();
    }
  }

  get [template]() {
    return templateFrom.html`
      <style>
        ${common}
        ${css}
      </style>
      ${html}
    `;
  }
}

/**
 * Toggles a boolean attribute. Differs from element.toggleAttribute()
 * by checking truthiness and setting boolean attributes only.
 * @param {object} element
 * @param {string} name
 * @param {boolean} value
 */
function toggleBooleanAttribute(element, name, value) {
  if (value) {
    element.setAttribute(name, '');
  } else {
    element.removeAttribute(name);
  }
}

/**
 * Move an attribute from one object to another.
 * @param {object} from
 * @param {object} to
 * @param {string} name
 * @param {*} value
 */
function delegateAttribute(from, to, name, value) {
  if (value || typeof value === 'boolean') {
    to.setAttribute(name, value);
    from.removeAttribute(name);
  } else {
    to.removeAttribute(name);
  }
}

/**
 * Dispatches a custom event that namespaces the event name to the
 * component's namespace.
 * @param {object} element
 * @param {string} name
 * @param {object} options
 */
function dispatchCustomEvent(element, name, options) {
  const namespace = element.tagName.split('-')[0].toLowerCase();
  const namespaceName = namespace + '-' + name;
  const customEvent = new CustomEvent(namespaceName, options);
  element.dispatchEvent(customEvent);
  return customEvent;
}
